//////////////////////////////////////////////////////////////////////////
// EasyX : A very simple DirectX Interface (for DOS-like programmers)   //
//                                                                      //
// Created by "Serkan YAZICI", 4 October 1997, Manchester               //
// Completely Freeware Source Code. Use It, Modify It as you wish       //
//                                                                      //
// Comments, Bugs, Credits welcomed...                                  //
//                   mchu4sy1@fs2.ee.umist.ac.uk                        //
//                                                                      //
//                           VERSION 1.0                                //
//////////////////////////////////////////////////////////////////////////


#include "EasyX.hpp"


//-----------------------------GLOBALS------------------------------------//


HINSTANCE hInst=0;         //handle of current app instance
HWND      appWnd=0;        //handle of app main window
BYTE      is_active=1;     //is the app active?
BYTE      *surface=0;      //linear surface to the graphics screen (backbuffer)


//-----------------------------VDU CLASS DEFINITION-----------------------//


LPDIRECTDRAW        VDU::ddraw   = NULL;
LPDIRECTDRAWSURFACE VDU::dsurface= NULL;
DDSURFACEDESC       VDU::ddsd;


void VDU::UpdateAll(void)
{
  if(dsurface==NULL)
    return;

  if(Lock())
    return;
  DWORD  i=0;
  DWORD *scr=(DWORD*)ddsd.lpSurface, *sur=(DWORD*)surface;
  while(i<76800){
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
    scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++; scr[i]=sur[i]; i++;
  }
  UnLock();
}


void VDU::Update(X1Y1_X2Y2)
{
  if(dsurface==NULL)
    return;

  if(Lock())
    return;
  register BYTE *screen=(BYTE*)ddsd.lpSurface;
  register short j,w=x2-x1+1;
  register DWORD t=((DWORD)y1<<9)+((DWORD)y1<<7)+x1;
  for(j=y1; j<=y2; j++){
    memcpy(screen+t,surface+t,w);
    t+=640;
  }
  UnLock();
}


void VDU::PutPixel(XY,BYTE col)
{
  if(dsurface==NULL)
    return;

  if(Lock())
    return;
  register BYTE *screen=(BYTE*)ddsd.lpSurface;
  screen[(y<<9)+(y<<7)+x] = col;
  UnLock();
}


BYTE VDU::GetPixel(XY)
{
  if(dsurface==NULL)
    return 0;

  if(LockRead())
    return 0;
  register BYTE *screen=(BYTE*)ddsd.lpSurface;
  register BYTE val = screen[(y<<9)+(y<<7)+x];
  UnLock();
  return val;
}


char VDU::Init(void)
{
  if(DD_OK != DirectDrawCreate(NULL,&ddraw,NULL))
    return 21;   //ERROR: DirectDraw Object could not be created
  if(DD_OK != ddraw->SetCooperativeLevel(appWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN))
    return 22;   //ERROR: Exclusive mode could not be set
  if(DD_OK != ddraw->SetDisplayMode(XLEN, YLEN, 8))
    return 23;   //ERROR: Graphics mode could not be changed
    
  ddsd.dwSize            = sizeof(ddsd);
  ddsd.dwFlags           = DDSD_CAPS|DDSD_BACKBUFFERCOUNT;
  ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
  ddsd.dwBackBufferCount = 1;
  if(DD_OK != ddraw->CreateSurface(&ddsd,&dsurface,NULL))
    return 24;   //ERROR: Primary surface could not be crated

  BYTE res=PALETTE::Init();
  if(res)
    return res;  //ERROR: Palette could not be initialised
  
  return 0;  /*NO ERROR*/
}


void VDU::Restore(void)
{
  if(ddraw != NULL){
    ddraw->RestoreDisplayMode();
    if(dsurface != NULL){
      dsurface->Release();
      dsurface = NULL;
    }
    ddraw->Release();
    ddraw = NULL;
  }
}


//-------------------------PALETTE CLASS DEFINITION-----------------------//


LPDIRECTDRAWPALETTE PALETTE::dpal;


BYTE PALETTE::Init(void)
{
  PALETTEENTRY lpColorTable[256];

  for(short i=0; i<256; i++){
    lpColorTable[i].peRed   = i;
    lpColorTable[i].peGreen = i;
    lpColorTable[i].peBlue  = i;
    lpColorTable[i].peFlags = 0;
  }

  if(DD_OK != VDU::ddraw->CreatePalette(DDPCAPS_8BIT|DDPCAPS_ALLOW256,lpColorTable,&dpal,NULL))
    return 25;  //ERROR: Could not create palette
    
  if(DD_OK != VDU::dsurface->SetPalette(dpal))
    return 26;  //ERROR: Could not set (associate) palette

  return 0;
}


inline void PALETTE::WaitRetrace(void)
{
  //VDU::ddraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
  VDU::ddraw->WaitForVerticalBlank(DDWAITVB_BLOCKEND,0);
}


//---------------------------TIMER CLASS DEFINITION-----------------------//


void (*TIMER::SecondaryTimer)(void)=0;     //Secondary Timer func.


void CALLBACK TIMER::Timer(HWND,UINT,UINT,DWORD)
{
  if(is_active){
    if(SecondaryTimer)
      SecondaryTimer();
  }
  else{
    MSG msg;
    if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)){
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }
}    


char TIMER::Install(void)
{
  SecondaryTimer = 0;
  if(!SetTimer(appWnd,1,50,(TIMERPROC)Timer))
    return 27;
  return 0;
}


void TIMER::Wait(DWORD ticks)
{
  DWORD current=Clock();
  while(abs(Clock()-current)<ticks)
    IDLE;
}


//-----------------------------KEY CLASS DEFINITION-----------------------//


short KEY::key=0;
short KEY::ch=0;
BYTE  KEY::status=0;


//----------------------------DRAW CLASS DEFINITION-----------------------//


inline Swap(short &a,short &b)
{
  short temp=a;
  a=b;
  b=temp;
}


void DRAW::VLine(XY,short h,BYTE color)
{
  register DWORD addr = y*XLEN + x;
  register DWORD end  = addr + h*XLEN;  
  for(register DWORD off=addr; off<end; off+=XLEN)
    surface[off] = color;
}


void DRAW::Line(X1Y1_X2Y2,BYTE color) //Draws a line using simplified Bresenham Algorithm
{
  register short d, dx=abs(x2-x1) ,dy=abs(y2-y1) ,Eincr ,NEincr ,yincr;

  if(dy <= dx){
    if(x2 < x1) { Swap(x1,x2); Swap(y1,y2); }
    if(y2 > y1) yincr = 1; else yincr = -1;
    d = (dy<<1) - dx;            // Initial decision variable value
    Eincr = dy<<1;               // Increment to move to E pixel
    NEincr = (dy - dx)<<1;       // Increment to move to NE pixel
    for(; x1 <= x2; x1++){
      Pixel(x1,y1,color);
      if(d < 0)
        d += Eincr;               // Choose the Eastern Pixel
      else{
        d += NEincr;              // Choose the North Eastern Pixel
        y1 += yincr;              // (or SE pixel for dx/dy < 0!)
      }
    }
  }
  else{
    if(y2 < y1) { Swap(x1,x2); Swap(y1,y2); }
    if(x2 > x1) yincr = 1; else yincr = -1;
    d = (dx<<1) - dy;             // Initial decision variable value
    Eincr = dx<<1;                // Increment to move to E pixel
    NEincr = (dx - dy)<<1;        // Increment to move to NE pixel
    for(; y1 <= y2; y1++){
      Pixel(x1,y1,color);
      if(d < 0)
        d += Eincr;               // Choose the Eastern Pixel
      else{
        d += NEincr;              // Choose the North Eastern Pixel
        x1 += yincr;              // (or SE pixel for dx/dy < 0!)
      }
    }
  }
}


void DRAW::FillRect(X1Y1_X2Y2,BYTE color)
{
  register short w=x2-x1+1;
  for(register short i=y1; i<=y2; i++)
    HLine(x1,i,w,color);
}


//--------------------------SYSTEM CLASS DEFINITION-----------------------//


short SYSTEM::Init(HINSTANCE hPrevInst,int cm)
{
  surface = (BYTE*)malloc(XLEN*YLEN+1024);
  if(!surface)
    return 1;
  memset(surface,0,XLEN*YLEN+1024);
    
  WNDCLASS wndclass;                 //CREATE MAIN WINDOW
  if(!hPrevInst){
    wndclass.style         = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc   = WndProc;
    wndclass.cbClsExtra    = wndclass.cbWndExtra  = 0;
    wndclass.hInstance     = hInst;
    wndclass.hIcon         = NULL;
    wndclass.hCursor       = LoadCursor(NULL,IDC_ARROW);//NULL;
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = MYINFO;
    RegisterClass(&wndclass);
  }
  appWnd = CreateWindow(MYINFO,MYINFO,WS_POPUP,0,0,GetSystemMetrics(SM_CXSCREEN),
                        GetSystemMetrics(SM_CYSCREEN),NULL,NULL,hInst,NULL);
  ShowWindow(appWnd,cm);
  UpdateWindow(appWnd);
  if(VDU::Init())                  //INIT & SET GRAPHICS (DirectX)
    return 2;

  if(TIMER::Install())
    return 3;
    
  return 0;  /*NO ERROR*/
}


void SYSTEM::Done(void)
{
  //***Restoring system***
  TIMER::Restore();
  VDU::Restore();
  free(surface);
}



int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE hPrevInst,LPSTR,int cm)
{
  hInst = hInstance;
  
  if(SYSTEM::Init(hPrevInst,cm)){
    MessageBox(NULL,"Error Initialising Application","Now Exiting...",MB_OK);
    if(appWnd)
      SendMessage(appWnd,WM_CLOSE,0,0);
  }
  else{
    Main();
    SYSTEM::Done();
    if(appWnd)
      SendMessage(appWnd,WM_CLOSE,0,0);
  }
  
  if(appWnd){
    MSG  msg;
    if(PeekMessage(&msg,NULL,0,0,PM_NOREMOVE)){
      while(GetMessage(&msg,NULL,0,0)){
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
  }
  
  return 0;
}


void DirectXIdle(void)
{
  MSG msg;
  if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)){
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
}


long PASCAL __export WndProc(HWND hwnd,UINT message,UINT wParam,LONG lParam)
{
  switch(message){
    case WM_ACTIVATE:
      {
        if(LOWORD(wParam)!=WA_INACTIVE)
          is_active = 1;
        else
          is_active = 0;
      }
      return 0;

    case WM_PAINT:
      {
        PAINTSTRUCT ps;
        HDC hdc = BeginPaint(hwnd,&ps);
        VDU::Update(ps.rcPaint.left,ps.rcPaint.top,ps.rcPaint.right,ps.rcPaint.bottom);
        EndPaint(hwnd,&ps);
      }
      return 0;

    case WM_CHAR:
      KEY::ch = (short)wParam;
      return 0;
      
    case WM_KEYDOWN: case WM_KEYUP: case WM_SYSKEYDOWN: case WM_SYSKEYUP:
      {
         KEY::key = (short)((lParam>>16)&0xFF);
         if(KEY::key==42 || KEY::key==54){  //shift
           if(message==WM_KEYDOWN) KEY::status|=1; else KEY::status&=0xFE;
         }
         if(KEY::key==29){                  //ctrl
           if(message==WM_KEYDOWN) KEY::status|=2; else KEY::status&=0xFD;
         }
         if(KEY::key==56){                  //alt
           if(message==WM_KEYDOWN || message==WM_SYSKEYDOWN) KEY::status|=4; else KEY::status&=0xFB;
         }
         if(message==WM_KEYUP)
           KEY::key += 128;
         else if(wParam>111 && wParam<122)
           KEY::ch = (short)wParam + 202;
      }
      break;

    case WM_CLOSE:
      DestroyWindow(hwnd);
      return 0;
        
    case WM_DESTROY:
      PostQuitMessage(0);
      return 0;
  }
  return DefWindowProc(hwnd,message,wParam,lParam);
}




