//////////////////////////////////////////////////////////////////////////
// EasyX : A very simple DirectX Interface (for DOS-like programmers)   //
//                                                                      //
// Created by "Serkan YAZICI", 4 October 1997, Manchester               //
// Completely Freeware Source Code. Use It, Modify It as you wish       //
//                                                                      //
// Comments, Bugs, Credits welcomed...                                  //
//                   mchu4sy1@fs2.ee.umist.ac.uk                        //
//                                                                      //
//                           VERSION 1.0                                //
//////////////////////////////////////////////////////////////////////////


#ifndef __EASYX_HPP
#define __EASYX_HPP


//----------------------------GENERAL STUFF-----------------------------//


extern "C" {              
  #include <stdio.h>            
  #include <stdlib.h>           
  #include <string.h>           
  #include <stdarg.h>           
  #include <ctype.h>            
  #include <windows.h>
  #include "ddraw.h"
}


typedef unsigned char  BYTE;         //8-bit unsigned
typedef signed char    SCHAR;        //8-bit signed
typedef unsigned short WORD;         //16-bit unsigned
typedef signed short   SHORT;        //16-bit signed
typedef unsigned long  DWORD;        //32-bit unsigned


#define MYINFO         "EasyX [Simplified DirectX Interface] Serkan YAZICI, 1997"
#define IDLE           DirectXIdle() //To be called in your input loops

#define XLEN           640           //X-Resolution
#define YLEN           480           //Y-Resolution


#define X1Y1_X2Y2      short x1,short y1,short x2,short y2   //2 points
#define X1Y1           short x1,short y1                     //1 point
#define XY             short x,short y                       //1 point


//-----------------------------GLOBALS------------------------------------//


extern HINSTANCE hInst;         //handle of current app instance
extern HWND      appWnd;        //handle of app main window
extern BYTE      is_active;     //is the app active?
extern BYTE      *surface;      //linear surface to the graphics screen (backbuffer)


//-----------------------FUNCTION PROTOTYPES------------------------------//


long PASCAL __export WndProc(HWND,UINT,UINT,LONG);  //App Message Handler
void DirectXIdle(void);                             //To be polled continuously
void Main(void);                                    //Your DOS-like main entry function


//---------------------------CLASSES--------------------------------------//


class SYSTEM {
public:
  static short Init(HINSTANCE,int);
  static void Done(void);
};


class VDU {
  friend class PALETTE;
  static LPDIRECTDRAW        ddraw;
  static LPDIRECTDRAWSURFACE dsurface;
  static DDSURFACEDESC       ddsd;
  static char Lock(void);
  static char LockRead(void);
  static void UnLock(void);
public:
  static char Init(void);
  static void Restore(void);
  
  static void UpdateAll(void);
  static void Update(X1Y1_X2Y2);
  static void PutPixel(XY,BYTE col);
  static BYTE GetPixel(XY);
};


class PALETTE {
  static LPDIRECTDRAWPALETTE dpal;
  static void WaitRetrace(void);
  static void Fader(void);
  static void AntiFader(void);
public:
  static BYTE Init(void);
  static void SetReg(BYTE i,BYTE red,BYTE green,BYTE blue);
  static void GetReg(BYTE i,BYTE &red,BYTE &green,BYTE &blue);
};


class TIMER {
  static void (*SecondaryTimer)(void);     
  static void CALLBACK Timer(HWND,UINT,UINT,DWORD);
public:
  static char Install(void);
  static void Wait(DWORD ticks);
  static DWORD Clock(void)                       { return timeGetTime()/55; }
  static void Restore(void)                      { KillTimer(appWnd,1); }
  static void InstallSecondary(void (*fn)(void)) { SecondaryTimer = fn; }
  static void RemoveSecondary(void)              { SecondaryTimer = 0;  }
};


class KEY {
  friend long PASCAL __export WndProc(HWND,UINT,UINT,LONG);
  static short key;
  static short ch;
  static BYTE  status;
public:
  static void Install(void)  { key=ch=status=0; }
  static void Restore(void)  { key=ch=status=0; }
  static BYTE ShiftOn(void)  { return BYTE(status&1); }
  static BYTE CtrlOn(void)   { return BYTE(status&2); }
  static BYTE AltOn(void)    { return BYTE(status&4); }
  static short Grab(void)    { return key; }
  static short Get(void)     { short c=ch; ch=0; return c;  }
  static short Scan(void)    { short k=key; key=0; return k;  }
};


class DRAW {
public:
  static BYTE GetPixel(XY);
  static void XORPixel(XY,BYTE col);
  static void Pixel(XY,BYTE col);
  static void HLine(XY,short w,BYTE col);
  static void VLine(XY,short h,BYTE col);
  static void Line(X1Y1_X2Y2,BYTE col);
  static void FillRect(X1Y1_X2Y2,BYTE col);
  static void Rectangle(X1Y1_X2Y2,BYTE col);
};


//---------------------------INLINE FUNCTIONS-----------------------------//


inline void PALETTE::SetReg(BYTE i,BYTE red,BYTE green,BYTE blue)
{
  PALETTEENTRY Rgb;
  Rgb.peRed   = ((red+1)<<2)-1;
  Rgb.peGreen = ((green+1)<<2)-1;
  Rgb.peBlue  = ((blue+1)<<2)-1;
  Rgb.peFlags = 0;
  dpal->SetEntries(0,i,1,&Rgb);
}


inline void PALETTE::GetReg(BYTE i,BYTE &red,BYTE &green,BYTE &blue)
{
  PALETTEENTRY Rgb;
  dpal->GetEntries(0,i,1,&Rgb);
  red   = ((short)Rgb.peRed+1)/4 -1;
  green = ((short)Rgb.peGreen+1)/4 -1;
  blue  = ((short)Rgb.peBlue+1)/4 -1;
}


inline char VDU::Lock(void)
{
  register HRESULT hr;
  register short   i=0;
  RECT     r;
  
  r.top = r.left = 0;
  r.right        = 639;
  r.bottom       = 479;
  do{
    hr = dsurface->Lock(&r,&ddsd,DDLOCK_WAIT|DDLOCK_WRITEONLY,0);
    if(hr==DDERR_SURFACELOST)
      dsurface->Restore();
    i++; if(i>1000) return 1;
  }while(hr!=DD_OK);
  return 0;
}


inline char VDU::LockRead(void)
{
  register HRESULT hr;
  register short   i=0;
  RECT     r;
  
  r.top = r.left = 0;
  r.right        = 639;
  r.bottom       = 479;
  do{
    hr = dsurface->Lock(&r,&ddsd,DDLOCK_WAIT|DDLOCK_READONLY,0);
    if(hr==DDERR_SURFACELOST)
      dsurface->Restore();
    i++; if(i>1000) return 1;
  }while(hr!=DD_OK);
  return 0;
}


inline void VDU::UnLock(void)
{
  register HRESULT hr;
  register short   i=0;
  
  do{
    hr = dsurface->Unlock(ddsd.lpSurface);
    i++; if(i>1000)return;
  }while(hr!=DD_OK);
}


inline void DRAW::Pixel(XY,BYTE col)            { surface[y*XLEN+x]=col; }
inline BYTE DRAW::GetPixel(XY)                  { return surface[y*XLEN+x]; }
inline void DRAW::XORPixel(XY,BYTE col)         { surface[y*XLEN+x] ^= col; }
inline void DRAW::HLine(XY,short _w,BYTE col)   { memset(surface+y*XLEN+x,col,_w); }


inline void DRAW::Rectangle(X1Y1_X2Y2,BYTE color)
{
  HLine(x1,y1,x2-x1+1,color);
  HLine(x1,y2,x2-x1+1,color);
  VLine(x1,y1,y2-y1+1,color);
  VLine(x2,y1,y2-y1+1,color);
}


#endif   /*__EASYX_HPP*/

